/*
Copyright (C) 2000 by Sean David Fleming

s.fleming@wbmt.tudelft.nl

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

The GNU GPL can also be found at http://www.gnu.org
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/times.h>

#include "gdis.h"

#define DEBUG 0

extern struct sysenv_pak sysenv;
extern struct elem_pak elements[];

gint selected_label = -1;
gint current_colour[3];
GtkWidget *label_list;

/***************************************/
/* apply PBC to the atoms in the model */
/***************************************/
void pbc_constrain_atoms(struct model_pak *data)
{
gint i, j, mov[3];
gfloat vec[3], lim;

g_return_if_fail(data != NULL);
/* requires bond & mol calc to have been done */

/* NOTES - non-orthogonal axes transformations are */
/* all taken care of in coord update routines, so */
/* that adding std xlat vec's to x,y,z should work */

/* TODO - if PBC transform is off - revert back to initial coords? */
/* molecule PBC translation - avoid breaking bonds */
/* NOTE - use original coords for PBC translation */
/* as the code will not work for rotated coords */
/* this unfortunately makes it non-reversible */
for (i=data->num_atoms ; i-- ; )
  {
/* don't do periodic images */
  if (!(data->atoms+i)->orig)
    {
    (data->atoms+i)->status = DELETED;
    continue;
    }
/* get coords */
/* NEW - should always be untransformed (ie scaled fractional) */
  vec[0] = (data->atoms+i)->x;
  vec[1] = (data->atoms+i)->y;
  vec[2] = (data->atoms+i)->z;
/* constrain */
  mov[0] = mov[1] = mov[2] = 0;
  for (j=data->periodic ; j-- ; )
    {
    if (data->pbc[j] < 0.0)
      vec[j] *= -1.0;
    lim = fabs(data->pbc[j]);
    while (vec[j] < 0.0)
      {
      vec[j] += lim;
      mov[j]++;
      }
    while (vec[j] > lim)
      {
      vec[j] -= lim;
      mov[j]--;
      }
    }

/* calculate the translation & apply */
  for (j=3 ; j-- ; )
    vec[j] = (gfloat) mov[j] * data->pbc[j];

  (data->atoms+i)->x += vec[0];
  (data->atoms+i)->y += vec[1];
  (data->atoms+i)->z += vec[2];

/* also to any attached shell */
  if ((data->atoms+i)->has_shell)
    {
    j = (data->atoms+i)->idx_shell;
    (data->shells+j)->x += vec[0];
    (data->shells+j)->y += vec[1];
    (data->shells+j)->z += vec[2];
    }
  }
mem_shrink(data);

/* NEW - delete repetitions */
/* These will happen if complete mols was used */
/* on a model with continuous bond chains */
/* FIXME - this will delete atoms that have partial occupancies!!! */
for (i=0 ; i<data->num_atoms-1 ; i++)
  {
  if ((data->atoms+i)->status == DELETED)
    continue;
  for (j=i+1 ; j<data->num_atoms ; j++)
    {
    vec[0] = (data->atoms+i)->x  - (data->atoms+j)->x;
    vec[1] = (data->atoms+i)->y  - (data->atoms+j)->y;
    vec[2] = (data->atoms+i)->z  - (data->atoms+j)->z;
    if (VEC3MAGSQ(vec) < POSITION_TOLERANCE)
      {
/* only delete if same atom type & no site occ. factor */
      if ((data->atoms+i)->atom_code == (data->atoms+j)->atom_code)
        if ((data->atoms+i)->sof > 0.9)
          if ((data->atoms+j)->sof > 0.9)
            (data->atoms+j)->status = DELETED;
      }
    }
  }
mem_shrink(data);
}

/**************************************/
/* apply PBC to the mols in the model */
/**************************************/
/* NB: assumes all mols are complete ie unfragmented */
/* unfragment() is very slow, this routine is faster and preferred */
/* for animations but assumes molecules are unbroken */
/* FIXME - is there a better way? (see main TODO) */
#define DEBUG_CONFINE_MOLS 0
void pbc_constrain_mols(struct model_pak *data)
{
gint i, j, k, m, mov[3];
gfloat vec[3], lim;

g_return_if_fail(data != NULL);

for (i=data->num_mols ; i-- ; )
  {
/* get coords (should be scaled untransformed fractional) */
  ARR3SET(vec, (data->mols+i)->centroid);

/* constrain */
  mov[0] = mov[1] = mov[2] = 0;
  for (j=data->periodic ; j-- ; )
    {
    if (data->pbc[j] < 0.0)
      vec[j] *= -1.0;
    lim = fabs(data->pbc[j]);
    while (vec[j] < 0.0)
      {
      vec[j] += lim;
      mov[j]++;
      }
    while (vec[j] > lim)
      {
      vec[j] -= lim;
      mov[j]--;
      }
    }

/* calculate the translation */
  for (j=3 ; j-- ; )
    vec[j] = (gfloat) mov[j] * data->pbc[j];

  for (j=3 ; j-- ; )
    (data->mols+i)->centroid[j] += vec[j];

#if DEBUG_CONFINE_MOLS
P3VEC("new: ",(data->mols+i)->centroid);
#endif

/* apply to all atoms in this molecule */
  for (j=(data->mols+i)->num_atoms ; j-- ; )
    {
    k = *((data->mols+i)->atom+j);
    (data->atoms+k)->x += vec[0];
    (data->atoms+k)->y += vec[1];
    (data->atoms+k)->z += vec[2];
/* also to any attached shell */
  if ((data->atoms+k)->has_shell)
    {
    m = (data->atoms+k)->idx_shell;
    (data->shells+m)->x += vec[0];
    (data->shells+m)->y += vec[1];
    (data->shells+m)->z += vec[2];
    }
    }
  }
}

/*******************************************/
/* draw a box - contents are the selection */
/*******************************************/
void update_box(gint x, gint y, struct model_pak *data, gint call_type)
{
switch (call_type)
  {
  case START:
/* setup the box selection object */
    data->box_on = TRUE;
    data->box_select[0].px = x;
    data->box_select[0].py = y;
    data->box_select[1].px = x;
    data->box_select[1].py = y;
    break;
  case UPDATE:
    data->box_select[1].px = x;
    data->box_select[1].py = y;
    break;
  }
}

/*************************************************/
/* find & highlight atoms within (box) selection */
/*************************************************/
void select_box()
{
gint x1,y1,x2,y2,tmp,i;
struct model_pak *data;

data = model_ptr(sysenv.active, RECALL);

x1 = data->box_select[0].px;
y1 = data->box_select[0].py;
x2 = data->box_select[1].px;
y2 = data->box_select[1].py;

/* ensure x1 < x2 and y1 < y2 */
if (x1 > x2)
  {
  tmp = x1;
  x1 = x2;
  x2 = tmp;
  }
if (y1 > y2)
  {
  tmp = y1;
  y1 = y2;
  y2 = tmp;
  }

/* get active subwin */
tmp = -1;
for (i=0 ; i<sysenv.num_displayed ; i++)
  if (sysenv.displayed[i] == sysenv.active)
    tmp = i;
/* subtract off subwin centre */
x1 -= sysenv.subcenx[tmp];
y1 -= sysenv.subceny[tmp];
x2 -= sysenv.subcenx[tmp];
y2 -= sysenv.subceny[tmp];

/* account for translation */
x1 -= data->offset[0];
y1 -= data->offset[1];
x2 -= data->offset[0];
y2 -= data->offset[1];

/* delete old selection (if any) */
clear_select();
/* compare and select if... */
for (i=0 ; i<data->num_atoms ; i++)
  {
/* ... not deleted */
  if ((data->atoms+i)->status == DELETED)
    continue;
/* ... and within box */
  if ((data->atoms+i)->px > x1 && (data->atoms+i)->px < x2)
    if ((data->atoms+i)->py > y1 && (data->atoms+i)->py < y2)
      {
/* add atom to the selection */
      if (data->select_size >= data->select_limit)
        mem_grow(data, SELECT, DEFAULT_SS);
      *(data->select_list+data->select_size) = i;
      data->select_size++;
/* highlight, better way? */
      (data->atoms+i)->status = SLITE;
      }
  }

/* done */
wire_draw(SINGLE);
}

/****************************************************/
/* create a general purpose colour selection dialog */
/****************************************************/
void colsel_dialog(GtkWidget *w, gpointer *obj)
{
gint id;
GtkWidget *csd;
GtkWidget *colsel;

/* check */
g_return_if_fail(obj != NULL);

/* create colour selection dialog */
csd = gtk_color_selection_dialog_new("Select color");
gtk_signal_connect(GTK_OBJECT(csd), "destroy",
                   GTK_SIGNAL_FUNC(gtk_widget_destroy), (gpointer) csd);
gtk_signal_connect_object 
       (GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(csd)->cancel_button),
       "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), (gpointer) csd);

/* TODO - move selection cursor to whatever the current colour is */
colsel = GTK_COLOR_SELECTION_DIALOG(csd)->colorsel;

/* callback routine depends on who called us */
id = GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(obj), "id"));
switch(id)
  {
  case MORPH_COLOUR:
/* set title */
/* apply colour change event */
    gtk_object_set_data(GTK_OBJECT(colsel), "id", (gpointer) MORPH_COLOUR);
    gtk_object_set_data(GTK_OBJECT(colsel), "ptr", (gpointer) 
                        gtk_object_get_data(GTK_OBJECT(obj), "ptr"));
    gtk_signal_connect 
       (GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(csd)->ok_button),
       "clicked", GTK_SIGNAL_FUNC(event_render_modify), (gpointer) colsel);
/* if continuous update */
/*
    gtk_signal_connect(GTK_OBJECT(colsel), "color_changed",
                       GTK_SIGNAL_FUNC(event_render_modify), (gpointer) colsel);
*/
    break;
  case ELEMENT_COLOUR:
/* apply colour change event */
    gtk_object_set_data(GTK_OBJECT(colsel), "wid", (gpointer) obj);
    gtk_object_set_data(GTK_OBJECT(colsel), "ptr", (gpointer) 
                        gtk_object_get_data(GTK_OBJECT(obj), "ptr"));
    gtk_signal_connect 
       (GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(csd)->ok_button),
       "clicked", GTK_SIGNAL_FUNC(elem_colour_change), (gpointer) colsel);
    break;
  }

gtk_widget_show(csd);
}

/*******************************************/
/* scan and delete invalid geometry labels */
/*******************************************/
#define DEBUG_CHECK_GEOM 0
void check_geom(struct model_pak *data)
{
gint i, j, n, flag, items;
struct geom_pak *ptr;

g_return_if_fail(data != NULL);

#if DEBUG_CHECK_GEOM
printf("Num old labels: %d\n",data->num_geom);
printf("Num atoms: %d\n",data->num_atoms);
#endif

n=0;
while (n<data->num_geom)
  {
  flag=0;
/* measurement type */
  items=2;
#if DEBUG_CHECK_GEOM
printf("label atoms: ");
#endif
  switch((data->geom+n)->type)
    {
    case TORSION:
      items++;
    case ANGLE:
      items++;
    case BOND:
    case DIST:
      ptr = data->geom + n;
      for (i=0 ; i<items ; i++)
        {
#if DEBUG_CHECK_GEOM
printf(" %d",*(ptr->list+i));
#endif
        if (*(ptr->list+i) >= data->num_atoms)
          {
          flag++;
/* FIXME - redundant code (see delete geom label) */
/* delete it */
          g_free(ptr->list);
          g_free(ptr->text);
/* FIXME - this shuffling'll cause problems with checking */
/* subsequent labels for validity */
          for (j=n ; j<data->num_geom-1 ; j++) 
            {
            (data->geom+j)->type = (data->geom+j+1)->type;
            (data->geom+j)->list = (data->geom+j+1)->list;
            (data->geom+j)->text = (data->geom+j+1)->text;
            }
/* update */
          data->num_geom--;
          break;
          }
        }
      break;
    }
  if (!flag)
    n++;
#if DEBUG_CHECK_GEOM
printf("\n");
#endif
  }
/* done */
#if DEBUG_CHECK_GEOM
printf("Num new labels: %d\n",data->num_geom);
#endif
}

/*********************************/
/* redo the geometry label clist */
/*********************************/
void update_geom_info()
{
gint i,j,k,l,n,flag;
gchar *info[3];
struct model_pak *data;

/* checks */
if (!sysenv.num_models)
  return;

flag=0;
for (i=0 ; i<MAX_DIALOGS ; i++)
  if (sysenv.dialog[i].type == GEOMETRY && sysenv.dialog[i].active)
    flag++;

if (!flag)
  return;

/* delete all existing entries */
gtk_clist_freeze((GtkCList *) label_list);
gtk_clist_clear((GtkCList *) label_list);

data = model_ptr(sysenv.active,RECALL);
if (data != NULL)
  {
  for (n=0 ; n<data->num_geom ; n++)
    {
/* measurement type */
    switch((data->geom+n)->type)
      {
      case BOND:
        info[0] = strdup("Bond");
/* convert # into atom labels */
        info[1] = (gchar *) g_malloc((2*ELEM_LABEL_SIZE+3)*sizeof(gchar));
        i = *((data->geom+n)->list);
        j = *((data->geom+n)->list+1);

/* TODO - else delete */
if (i < data->num_atoms && j < data->num_atoms)
  {

        g_snprintf(info[1],8,"%2s : %-2s",(data->atoms+i)->element
                                      ,(data->atoms+j)->element);
/* get from text label */
        info[2] = (data->geom+n)->text;
  }
        break;
      case DIST:
        info[0] = strdup("Dist");
        info[1] = (gchar *) g_malloc((2*ELEM_LABEL_SIZE+1)*sizeof(gchar));
        i = *((data->geom+n)->list);
        j = *((data->geom+n)->list+1);
        g_snprintf(info[1],8,"%2s : %-2s",(data->atoms+i)->element
                                      ,(data->atoms+j)->element);
        info[2] = (data->geom+n)->text;
        break;
      case ANGLE:
        info[0] = strdup("Angle");
/* convert # into atom labels */
        info[1] = (gchar *) g_malloc((3*ELEM_LABEL_SIZE+2)*sizeof(gchar));
        i = *((data->geom+n)->list);
        j = *((data->geom+n)->list+1);
        k = *((data->geom+n)->list+2);
        g_snprintf(info[1],13,"%2s : %-2s : %-2s",(data->atoms+i)->element
                                                ,(data->atoms+j)->element
                                                ,(data->atoms+k)->element);
/* get from text label */
        info[2] = (data->geom+n)->text;
        break;
      case TORSION:
        info[0] = strdup("Torsion");
/* convert # into atom labels */
        info[1] = (gchar *) g_malloc((4*ELEM_LABEL_SIZE+4)*sizeof(gchar));
        i = *((data->geom+n)->list);
        j = *((data->geom+n)->list+1);
        k = *((data->geom+n)->list+2);
        l = *((data->geom+n)->list+3);
        g_snprintf(info[1],18,"%2s : %-2s : %-2s : %-2s",
                               (data->atoms+i)->element,
                               (data->atoms+j)->element,
                               (data->atoms+k)->element,
                               (data->atoms+l)->element);
/* get from text label */
        info[2] = (data->geom+n)->text;
        break;
      default:
        info[0] = strdup("Unknown");
        info[1] = strdup("?");
        info[2] = NULL;
      }
/* add to the list */
    gtk_clist_append((GtkCList *) label_list, &info[0]);
/* free */
    g_free(info[0]);
    g_free(info[1]);
    } 
  }

/* update */
gtk_clist_thaw((GtkCList *) label_list);
}

/***************************/
/* geometry label creation */
/***************************/
gint geom_label(gint type, struct model_pak *data, gint *atoms, gchar *label)
{
gint n;

/* checks */
if (!data->geom_limit)
  return(1);

/* do we need more memory? */
while (data->num_geom >= data->geom_limit)
  mem_grow(data, GEOM, DEFAULT_NL);
n = data->num_geom;

(data->geom+n)->type = type;
switch(type)
  {
  case BOND:
  case DIST:
/* allocate */
    (data->geom+n)->list = (gint *) g_malloc(2*sizeof(gint));
/* assign */ 
    *((data->geom+n)->list) = *(atoms);
    *((data->geom+n)->list+1) = *(atoms+1);
    (data->geom+n)->text = strdup(label);
    break;
  case ANGLE:
/* allocate */
    (data->geom+n)->list = (gint *) g_malloc(3*sizeof(gint));
/* assign */ 
    ARR3SET((data->geom+n)->list, atoms);
    (data->geom+n)->text = strdup(label);
    break;
  case TORSION:
/* allocate */
    (data->geom+n)->list = (gint *) g_malloc(4*sizeof(gint));
/* assign */ 
    *((data->geom+n)->list) = *(atoms);
    *((data->geom+n)->list+1) = *(atoms+1);
    *((data->geom+n)->list+2) = *(atoms+2);
    *((data->geom+n)->list+3) = *(atoms+3);
    (data->geom+n)->text = strdup(label);
    break;
  default:
    printf("geom_label() error: unknown type.\n");
    return(1);
  }

/* creation successful */
data->num_geom++;
return(0);
}

/***************************/
/* geometry label deletion */
/***************************/
void delete_geom_label(gint mode)
{
gint i, n, flag;
struct model_pak *data;

/* checks */
if (!sysenv.num_models)
  return;

flag=0;
for (i=0 ; i<MAX_DIALOGS ; i++)
  if (sysenv.dialog[i].type == GEOMETRY && sysenv.dialog[i].active)
    flag++;

if (!flag)
  return;

data = model_ptr(sysenv.active, RECALL);
if (data == NULL)
  return;

/* perform the requested deletion */
switch(mode)
  {
  case SINGLE:
/* valid selection? */
    if (selected_label >= 0)
      {
/* delete it */
      g_free((data->geom+selected_label)->list);
      g_free((data->geom+selected_label)->text);
      for (n=selected_label ; n<data->num_geom-1 ; n++) 
        {
        (data->geom+n)->type = (data->geom+n+1)->type;
        (data->geom+n)->list = (data->geom+n+1)->list;
        (data->geom+n)->text = (data->geom+n+1)->text;
        }
/* update */
      data->num_geom--;
      if (selected_label == data->num_geom)
        selected_label--;
      }
    break;
  case ALL:
/* delete all labels */
    for (n=0 ; n<data->num_geom ; n++)
      {
      g_free((data->geom+n)->list);
      g_free((data->geom+n)->text);
      }
/* update */
    data->num_geom = 0;
    selected_label = -1;
    break;
  }

/* update & redraw */
update_geom_info();
wire_draw(SINGLE);
}

/*************/
/* find bond */
/*************/
gint seek_bond(gint subwin, gint x, gint y, struct model_pak *data)
{
gint n;
gfloat dx, dy, dr2;

/* subtract off subwin centre */
x -= sysenv.subcenx[subwin];
y -= sysenv.subceny[subwin];

for (n=0 ; n<data->num_bonds ; n++)
  {
  dx = ((data->bonds+n)->px - x) + data->offset[0]; 
  dy = ((data->bonds+n)->py - y) + data->offset[1]; 
  dr2 = dx*dx + dy*dy;
/* FIXME - too arbitrary? */
  if (dr2 < 36)
    return(n);
  }
return(-1);
}

/*****************/
/* info on bonds */
/*****************/
void info_bond(gint subwin, gint x, gint y, struct model_pak *data)
{
gint i,j,n,atom[2];
gfloat r,rx,ry,rz;
gchar txt[80];

/* seek */
n = seek_bond(subwin, x, y, data);
if (n < 0)
  return;

i = (data->bonds+n)->atom1;
j = (data->bonds+n)->atom2;
/* separation */
rx = (data->atoms+i)->rx - (data->atoms+j)->rx;
ry = (data->atoms+i)->ry - (data->atoms+j)->ry;
rz = (data->atoms+i)->rz - (data->atoms+j)->rz;
r = sqrt(rx*rx + ry*ry + rz*rz);
/* create geometry label */
atom[0] = i;
atom[1] = j;
g_snprintf(txt,8,"%7.3f",r);
geom_label(BOND, data, atom, txt);
update_geom_info();
/* redraw */
wire_draw(SINGLE);
}

/*************************/
/* bond overlay modifier */
/*************************/
void bond_overlay(struct model_pak *data)
{
gint i, j, m, n;
gint count;

/* search normal bonds */
for (n=0 ; n<data->num_bonds ; n++)
  {
/* retrieve constituents */
  i = (data->bonds+n)->atom1;
  j = (data->bonds+n)->atom2;
/* search user bonds... */
  for (m=0 ; m<data->num_ubonds ; m++)
    {
    count=0;
    if (i == (data->ubonds+m)->atom1)
      count++;
    if (i == (data->ubonds+m)->atom2)
      count++;
    if (j == (data->ubonds+m)->atom1)
      count++;
    if (j == (data->ubonds+m)->atom2)
      count++;
/* overlay the type */
    if (count > 1)
      {
      (data->bonds+n)->type = (data->ubonds+m)->type; 
      break;
      }
    }
  }
}

/* must be an existing (normal) bond */
/* this makes selection (one click) and drawing easier */
void user_bond(gint subwin, gint x, gint y, struct model_pak *data, gint type)
{
gint i, j, n;
gint count, flag;

/* valid bond? */
n = seek_bond(subwin, x, y, data);
if (n < 0)
  return;
/* retrieve constituents */
i = (data->bonds+n)->atom1;
j = (data->bonds+n)->atom2;

/* search... */
flag=0;
for (n=0 ; n<data->num_ubonds ; n++)
  {
  count=0;
  if (i == (data->ubonds+n)->atom1)
    count++;
  if (i == (data->ubonds+n)->atom2)
    count++;
  if (j == (data->ubonds+n)->atom1)
    count++;
  if (j == (data->ubonds+n)->atom2)
    count++;
/* ...and make the change */
  if (count > 1)
    {
    (data->ubonds+n)->type = type; 
    flag=1;
    break;
    }
  }

/* not found - create */
if (!flag)
  {
  n = data->num_ubonds;
  if (++data->num_ubonds >= data->ubond_limit)
    mem_grow(data, UBOND, 10);
  (data->ubonds+n)->atom1 = i;
  (data->ubonds+n)->atom2 = j;
  (data->ubonds+n)->type = type;
  }

/* done */
wire_draw(SINGLE);
}

/*****************/
/* info on dists */
/*****************/
void info_dist(gint x, gint y, struct model_pak *data)
{
static gint calls=0, first=0, last=0;
gint i, atom[2];
gfloat rx,ry,rz,r;
gchar txt[80];

/* (kludgy) fix for picking only one atom, switching modes */
/* and then changing back, and having the first new atom */
/* really the last atom of the previous time */
/* clean up call (mode switching) */
if (x < 0)
  {
  calls=first=last=0;
  return;
  }

/* attempt to match point with an atom */
i = seek_atom(x, y, data);

/* no match - leave */
if (i < 0)
  return;
else
  calls++;

/* are we looking at the 1st or the 2nd call? */
switch (calls)
  {
  case 1:
/* clear any previous highlighting */
/* should be ok on very 1st call - assuming no highlighting */
/* is allowed to persist after a mode change */
    if ((data->atoms+first)->status == SLITE)
      (data->atoms+first)->status = NORMAL;
/* save */
    first = i;
/* highlight selected atom */
    (data->atoms+first)->status = SLITE;
    wire_draw(SINGLE);
    break;
  case 2:
/* save */
    last = i;
/* remove highlight (we now have a label) */
    (data->atoms+first)->status = NORMAL;
/* we now have two atoms we can calculate a distance for */
    rx = (data->atoms+first)->rx - (data->atoms+i)->rx;
    ry = (data->atoms+first)->ry - (data->atoms+i)->ry;
    rz = (data->atoms+first)->rz - (data->atoms+i)->rz;
/* fractional ?*/
    r = sqrt(rx*rx + ry*ry + rz*rz);
/* create geometry label */
    atom[0] = first;
    atom[1] = last;
    g_snprintf(txt,8,"%7.3f",r);
    geom_label(DIST, data, atom, txt);
    update_geom_info();
/* redraw */
    wire_draw(SINGLE);
/* reset counter */
    calls=0;
    break;
  default:
    printf("Error: bad state in info_dist()\n");
    calls=0;
  }
}

/******************/
/* info on angles */
/******************/
void info_angle(gint x, gint y, struct model_pak *data)
{
static gint calls=0, atoms[3]={0,0,0};
gint i;
gfloat a[3], b[3], angle;
gchar txt[40];

/* (kludgy) fix for picking only one atom, switching modes */
/* and then changing back, and having the first new atom */
/* really the last atom of the previous time */
/* clean up call (mode switching) */
if (x < 0)
  {
  calls=atoms[0]=atoms[1]=atoms[2]=0;
  return;
  }

/* attempt to match point with an atom */
i = seek_atom(x, y, data);

/* no match - leave */
if (i < 0)
  return;
else
  calls++;

/* are we looking at the 1st or the 2nd call? */
switch (calls)
  {
  case 1:
/* clear any previous highlighting */
/* should be ok on very 1st call - assuming no highlighting */
/* is allowed to persist after a mode change */
    if ((data->atoms+atoms[0])->status == SLITE)
      (data->atoms+atoms[0])->status = NORMAL;
/* save */
    atoms[0] = i;
/* highlight selected atom */
    (data->atoms+i)->status = SLITE;
    break;
  case 2:
/* save */
    atoms[1] = i;
/* highlight selected atom */
    (data->atoms+i)->status = SLITE;
/*
    g_snprintf(txt,12,"Selected %-2s", (data->atoms+i)->element);
*/
    break;
  case 3:
/* we now have three atoms we can calculate an angle for */
    atoms[2] = i;
/* remove highlights (we now have a label) */
    (data->atoms+atoms[0])->status = NORMAL;
    (data->atoms+atoms[1])->status = NORMAL;
/* compute vectors (from center atom) */
    a[0] = (data->atoms+atoms[0])->rx - (data->atoms+atoms[1])->rx;
    a[1] = (data->atoms+atoms[0])->ry - (data->atoms+atoms[1])->ry;
    a[2] = (data->atoms+atoms[0])->rz - (data->atoms+atoms[1])->rz;
    b[0] = (data->atoms+atoms[2])->rx - (data->atoms+atoms[1])->rx;
    b[1] = (data->atoms+atoms[2])->ry - (data->atoms+atoms[1])->ry;
    b[2] = (data->atoms+atoms[2])->rz - (data->atoms+atoms[1])->rz;
    angle = 180.0*via(a,b,3)/PI; 
/* NEW - unified geometry label creation */
/* TODO - if atoms aren't bonded - draw a yellow line between them? */
/* eg a distance label with blank text as the value */
    g_snprintf(txt,8,"%7.3f",fabs(angle));
    geom_label(ANGLE, data, atoms, txt);
    update_geom_info();
/* reset counter */
    calls=0;
    break;
  default:
    printf("Error: bad state in info_angles()\n");
    calls=0;
    return;
  }
/* update */
wire_draw(SINGLE);
}

/****************************/
/* info on torsional angles */
/****************************/
void info_torsion(gint x, gint y, struct model_pak *data)
{
static gint calls=0, atoms[4]={0,0,0,0};
gint i;
gfloat a[3], b[3], n[4], angle, len;
gchar txt[40];

/* (kludgy) fix for picking only one atom, switching modes */
/* and then changing back, and having the first new atom */
/* really the last atom of the previous time */
/* clean up call (mode switching) */
if (x < 0)
  {
  calls=atoms[0]=atoms[1]=atoms[2]=atoms[3]=0;
  return;
  }

/* attempt to match point with an atom */
i = seek_atom(x, y, data);

/* no match - leave */
if (i < 0)
  return;
else
  calls++;

/* what stage are we at? */
switch (calls)
  {
  case 1:
/* FIXME - do we need to clear any previous highlighting? */
  case 2:
  case 3:
/* save */
    atoms[calls-1] = i;
/* highlight selected atom */
    (data->atoms+i)->status = SLITE;
    break;
  case 4:
/* we now have four atoms we can calculate a dihedral for */
    atoms[3] = i;
/* remove highlights (we now have a label) */
    for(i=0 ; i<3 ; i++)
      (data->atoms+atoms[i])->status = NORMAL;
/* compute end atom vectors (from 1-2 axis) */
    a[0] = (data->atoms+atoms[0])->rx - (data->atoms+atoms[1])->rx;
    a[1] = (data->atoms+atoms[0])->ry - (data->atoms+atoms[1])->ry;
    a[2] = (data->atoms+atoms[0])->rz - (data->atoms+atoms[1])->rz;
    b[0] = (data->atoms+atoms[3])->rx - (data->atoms+atoms[2])->rx;
    b[1] = (data->atoms+atoms[3])->ry - (data->atoms+atoms[2])->ry;
    b[2] = (data->atoms+atoms[3])->rz - (data->atoms+atoms[2])->rz;
/* compute normal to the plane in which the dihedral is to be calc'd */
    n[0] = (data->atoms+atoms[1])->rx - (data->atoms+atoms[2])->rx;
    n[1] = (data->atoms+atoms[1])->ry - (data->atoms+atoms[2])->ry;
    n[2] = (data->atoms+atoms[1])->rz - (data->atoms+atoms[2])->rz;
/* normalize */
    len = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]); 
    n[0] /= len;
    n[1] /= len;
    n[2] /= len;
/* n[0..2] is the normal & n[0..3] is the equation of the plane thru 2 */
/* project 1-0 onto the normal (ie dist to plane) */
    len = a[0]*n[0] + a[1]*n[1] + a[2]*n[2];
/* subtract vector to plane from 1-0 to get projection on plane */
    a[0] -= len*n[0];
    a[1] -= len*n[1];
    a[2] -= len*n[2];
/* project 2-3 onto the normal */
    len = b[0]*n[0] + b[1]*n[1] + b[2]*n[2];
/* subtract vector to plane from 2-3 to get projection on plane */
    b[0] -= len*n[0];
    b[1] -= len*n[1];
    b[2] -= len*n[2];
/* compute angle between projected vectors */
    angle = 180.0*via(a,b,3)/PI; 
/* NEW - unified geometry label creation */
/* TODO - if atoms aren't bonded - draw a yellow line between them? */
/* eg a distance label with blank text as the value */
    g_snprintf(txt,8,"%7.3f",fabs(angle));
    geom_label(TORSION, data, atoms, txt);
    update_geom_info();
/* reset counter */
    calls=0;
    break;
  default:
    printf("Error: bad state in info_angles()\n");
    calls=0;
    return;
  }
/* update */
wire_draw(SINGLE);
}

/**************************/
/* callbacks for the list */
/**************************/
void select_label(GtkWidget *clist, gint row, gint column,
                         GdkEventButton *event, gpointer data)
{
selected_label = row;
}
void unselect_label(GtkWidget *clist, gint row, gint column,
                         GdkEventButton *event, gpointer data)
{
selected_label = -1;
}

/**************************/
/* NEW - geometric labels */
/**************************/
void geom_info()
{
gint id;
GtkWidget *frame, *vbox, *hbox, *button;
GtkWidget *scr_win;
gchar *titles[3] = {"  Type  ", " Constituent atoms ", " Value "};
struct dialog_pak *geom_dialog;

/* already got a dialog? */
if (!sysenv.num_models)
  return;

/* request a new dialog */
if ((id = request_dialog(sysenv.active, GEOMETRY)) < 0)
  return;
geom_dialog = &sysenv.dialog[id];

/* create new dialog */
geom_dialog->win = gtk_dialog_new();
gtk_window_set_title(GTK_WINDOW (geom_dialog->win), "Geometry panel");
gtk_window_set_default_size(GTK_WINDOW(geom_dialog->win), 250, 400);
gtk_signal_connect(GTK_OBJECT(geom_dialog->win), "destroy",
                   GTK_SIGNAL_FUNC(event_close_dialog), (gpointer) id);

/* Frame 1 */
frame = gtk_frame_new ("Measurement type");
gtk_box_pack_start(GTK_BOX(GTK_DIALOG(geom_dialog->win)->vbox),frame,FALSE,TRUE,0); 
gtk_container_set_border_width (GTK_CONTAINER(frame), 5);

/* hbox for the modes */
hbox = gtk_hbox_new(TRUE, 0);
gtk_container_add(GTK_CONTAINER(frame), hbox);
gtk_container_set_border_width (GTK_CONTAINER(hbox), 5);

/* button 1 */
button = gtk_button_new_with_label ("Bond");
gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
gtk_signal_connect_object(GTK_OBJECT (button), "clicked",
                          GTK_SIGNAL_FUNC (switch_mode),
                         (gpointer) BOND_INFO);

/* button 2 */
button = gtk_button_new_with_label (" Distance ");
gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
gtk_signal_connect_object(GTK_OBJECT (button), "clicked",
                          GTK_SIGNAL_FUNC (switch_mode),
                         (gpointer) DIST_INFO);

/* button 3 */
button = gtk_button_new_with_label ("Angle");
gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
gtk_signal_connect_object(GTK_OBJECT (button), "clicked",
                          GTK_SIGNAL_FUNC (switch_mode),
                         (gpointer) ANGLE_INFO);
/* button 4 */
button = gtk_button_new_with_label (" Torsion ");
gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
gtk_signal_connect_object(GTK_OBJECT (button), "clicked",
                          GTK_SIGNAL_FUNC (switch_mode),
                         (gpointer) TORSION_INFO);

/* Frame 2 */
frame = gtk_frame_new ("Label list");
gtk_box_pack_start(GTK_BOX(GTK_DIALOG(geom_dialog->win)->vbox),frame,TRUE,TRUE,0); 
gtk_container_set_border_width (GTK_CONTAINER(frame), 5);
/* create a vbox in the frame */
vbox = gtk_vbox_new (FALSE, 0);
gtk_container_add (GTK_CONTAINER(frame), vbox);
gtk_container_set_border_width (GTK_CONTAINER(GTK_BOX(vbox)), 3);
/* scrolled model pane */
scr_win = gtk_scrolled_window_new (NULL, NULL);
gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scr_win),
                                GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
gtk_box_pack_start(GTK_BOX(vbox), scr_win, TRUE, TRUE, 0);
/* Create the CList */
label_list = gtk_clist_new_with_titles(3, titles);
/* events */
gtk_signal_connect(GTK_OBJECT(label_list), "select_row",
                   GTK_SIGNAL_FUNC(select_label), NULL);
gtk_signal_connect(GTK_OBJECT(label_list), "unselect_row",
                   GTK_SIGNAL_FUNC(unselect_label), NULL);

/* set some properties of the CList */
gtk_clist_set_shadow_type (GTK_CLIST(label_list), GTK_SHADOW_OUT);
gtk_container_add(GTK_CONTAINER(scr_win), label_list);

/* Delete buttons */
hbox = gtk_hbox_new(TRUE, 0);
gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);
/* col 1 - only one */
button = gtk_button_new_with_label ("  Delete  ");
gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
                           GTK_SIGNAL_FUNC (delete_geom_label),
                           (gpointer) SINGLE);
/* col 2 - all labels */
button = gtk_button_new_with_label (" Clear all ");
gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
                           GTK_SIGNAL_FUNC (delete_geom_label),
                           (gpointer) ALL);

/* terminating button */
button = gtk_button_new_with_label ("  Close  ");
gtk_box_pack_start (GTK_BOX (GTK_DIALOG (geom_dialog->win)->action_area),
                                         button, FALSE, FALSE, 0);
gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
                           GTK_SIGNAL_FUNC(close_dialog),
                           (gpointer) id);

/* done */
gtk_widget_show_all(geom_dialog->win);
/* refresh labels */
update_geom_info();
}

